#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <pthread.h>
#include <unistd.h>

#include <sys/stat.h>
#include <sys/types.h>

#include <profapi.h>

#include "config.h"
#include "blink.h"
#include "cmd.h"

#ifdef ENABLE_BLINK1
#include "blink1-lib.h"
#endif

#define THINKLIGHT_BASE "/sys/class/leds/tpacpi::thinklight/"
#define THINKLIGHT_MAX THINKLIGHT_BASE "max_brightness"
#define THINKLIGHT THINKLIGHT_BASE "brightness"
#define THINKLIGHT_DELAY 200000
#define THINKLIGHT_BLINKS 2
#define THINKLIGHT_ON 255
#define THINKLIGHT_OFF 0


static pthread_mutex_t lock;

static void
blink_noop(enum event event __attribute__((unused)))
{
	prof_log_error("No blinking device active");
}

struct blink_backend led = { "No LED found", blink_noop };

static void
blink_led(enum event event)
{
	int fd, i, prev;
	char buf[32];

	if (event == LED_OFF)
	{
		return;
	}

	fd = open(THINKLIGHT, O_RDWR);
	if (fd == -1)
	{
		prof_log_error("Failed to open ThinkLight device");
		return;
	}

	prev = 0;
	memset(buf, 0, sizeof(buf));

	if (read(fd, buf, sizeof(buf) - 1) == -1)
	{
		prof_log_error("Failed to read ThinkLight status");
		close(fd);
		return;
	}

	prev = atoi(buf);

	for (i = 0; i < THINKLIGHT_BLINKS; i++)
	{
		dprintf(fd, "%d", THINKLIGHT_ON);
		usleep(THINKLIGHT_DELAY);
		dprintf(fd, "%d", THINKLIGHT_OFF);
		usleep(THINKLIGHT_DELAY);
	}

	dprintf(fd, "%d", prev);
	close(fd);
}

#ifdef ENABLE_BLINK1
struct pattern
{
	uint16_t duration;
	uint8_t r;
	uint8_t g;
	uint8_t b;
};

/* patterns need to be terminated by a row of 0xff */
struct pattern pat_led_off[] =
{
	{ 100, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff }
};

struct pattern pat_init[] =
{
	{ 50, 0x00, 0x00, 0x00 },
	{ 100, 0x00, 0x00, 0xff },
	{ 100, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff }
};

struct pattern pat_msg[] =
{
	{ 50, 0x00, 0x00, 0x00 },
	{ 100, 0xff, 0xff, 0x00 },
	{ 100, 0x00, 0x00, 0x00 },
	{ 100, 0xff, 0xff, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff }
};

static void
play_pattern(blink1_device *dev, struct pattern *pat)
{
	unsigned int i;
	struct pattern *p;
	char buf[128];

	prof_log_debug("[blink] playing pattern:");

	for (i=0;; i++)
	{
		p = &pat[i];

		if (p->duration == 0xff && p->r == 0xff && p->g == 0xff &&
								p->b == 0xff)
		{
			prof_log_debug("[blink] pattern complete");
			break;
		}

		snprintf(buf, sizeof(buf), "[blink] - %d/%d/%d (%dms)", p->r,
						p->g, p->b, p->duration);
		prof_log_debug(buf);
		blink1_fadeToRGB(dev, p->duration, p->r, p->g, p->b);
		blink1_sleep(p->duration);
	}
}

static void
blink_blink1(enum event e)
{
	struct pattern *p;
	blink1_device *dev;

	switch (e)
	{
	case STARTUP:
		p = pat_init;
		break;
	case NEW_MSG:
		p = pat_msg;
		break;
	case LED_OFF:
		p = pat_led_off;
		break;
	default:
		p = pat_led_off;
		break;
	}

	if (blink1_enumerate() < 1)
	{
		prof_log_error("No blink devices found (2)");
		return;
	}

	dev = blink1_openById(0);
	if (dev == NULL)
	{
		prof_log_error("Failed to open device");
		return;
	}

	play_pattern(dev, p);
	blink1_close(dev);
}
#endif

static void *
do_blink(void *arg)
{
	enum event event;
	event = (enum event) arg;

	pthread_mutex_lock(&lock);
	led.blink(event);
	pthread_mutex_unlock(&lock);

	return NULL;
}

void
blink(enum event e)
{
	pthread_t t;

	prof_log_debug("[blink] blink");

	if (pthread_create(&t, NULL, do_blink, (void *) e) != 0)
	{
		prof_log_error("Failed to create thread");
		return;
	}
	pthread_detach(t);
}

void
prof_cons_error(const char *msg)
{
	prof_cons_show_themed("blink", "error", "red", msg);
	prof_cons_alert();
}

/*
 * Profanity callbacks
 */

static int
find_backend(struct blink_backend *backend)
{
	char msg[128], led_dev[128];

	snprintf(led_dev, sizeof(led_dev), "%s", THINKLIGHT);

#ifdef ENABLE_BLINK1
	if (blink1_enumerate() > 0)
	{
		prof_log_info("blink(1) LED found");
		prof_cons_show("Using blink(1) LED for notifications");
		snprintf(backend->name, sizeof(backend->name), "blink(1)");
		backend->blink = blink_blink1;
		return 0;
	}
#endif

	if (access(led_dev, W_OK) == 0)
	{
		prof_log_info("ThinkLight found");
		prof_cons_show("Using ThinkLight for notifications");
		snprintf(backend->name, sizeof(backend->name), "ThinkLight");
		backend->blink = blink_led;
		return 0;
	}
	else if (access(led_dev, R_OK) == 0)
	{
		snprintf(msg, sizeof(msg), "%s: %s", led_dev, strerror(errno));
		prof_cons_error(msg);
	}

	return 1;
}

void
prof_init(const char * const version __attribute__((unused)),
	const char * const status __attribute__((unused)),
	const char *const account_name __attribute__((unused)),
	const char *const fulljid __attribute__((unused)))
{
	char buf[64];
	prof_log_error("prof_on_init()");

	if (pthread_mutex_init(&lock, NULL) != 0)
	{
		prof_log_error("Failed to initialize mutex");
	}

	snprintf(buf, sizeof(buf), "blink plugin %s initializing", VERSION);
	prof_log_info(buf);

	if (find_backend(&led) != 0)
	{
		prof_cons_error("No blinking device found");
	}

	blink_cmd_init(&led);
	blink(STARTUP);
}

void
prof_on_shutdown()
{
	prof_log_debug("prof_on_shutdown()");
	pthread_mutex_destroy(&lock);
}

void
prof_on_unload()
{
	prof_log_debug("prof_on_unload()");
}

void
prof_on_chat_win_focus(const char *const barejid __attribute__((unused)))
{
	prof_log_debug("[blink] prof_on_chat_win_focus");
	blink(LED_OFF);
}

void
prof_on_room_win_focus(const char *const barejid __attribute__((unused)))
{
	prof_log_debug("[blink] prof_on_room_win_focus");
	blink(LED_OFF);
}

void
prof_post_priv_message_display(
	const char *const barejid __attribute__((unused)),
	const char *const nick __attribute__((unused)),
	const char *message __attribute__((unused))
)
{
	prof_log_debug("[blink] prof_post_priv_message_display");
	blink(NEW_MSG);
}

void
prof_post_room_message_display(
	const char *const barejid __attribute__((unused)),
	const char *const nick __attribute__((unused)),
	const char *message __attribute__((unused))
)
{
	prof_log_debug("[blink] prof_post_room_message_display");
	blink(NEW_MSG);
}

void
prof_post_chat_message_display(
	const char *const barejid __attribute__((unused)),
	const char *const nick __attribute__((unused)),
	const char *message __attribute__((unused))
)
{
	prof_log_debug("[blink] prof_post_chat_message_display");
	blink(NEW_MSG);
}
